/*
 * Copyright (C) 2017-2024 Dominic Heutelbeck (dominic@heutelbeck.com)
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.sapl.springdatacommon.handlers;

import static io.sapl.springdatacommon.sapl.utils.Utilities.CONDITIONS;
import static io.sapl.springdatacommon.sapl.utils.Utilities.TYPE;

import java.util.Objects;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.JsonNodeFactory;

/**
 * This class takes care of extracting the correct obligation for manipulating
 * the query.
 */
public class QueryManipulationObligationProvider {

    private static final ObjectMapper MAPPER    = new ObjectMapper();
    private static final JsonNode     NULL_NODE = JsonNodeFactory.instance.nullNode();

    /**
     * Extracts the query CONDITION of an obligation to apply the corresponding
     * QueryManipulation.
     *
     * @param obligation which contains query CONDITIONS.
     * @return all query CONDITIONS.
     */
    public ArrayNode getConditions(JsonNode obligation) {
        if (obligation.has(CONDITIONS) && obligation.get(CONDITIONS).isArray()
                && !obligation.get(CONDITIONS).isEmpty()) {
            return (ArrayNode) obligation.get(CONDITIONS);
        }
        return MAPPER.createArrayNode();
    }

    /**
     * Extracts the correct obligation from all obligations to apply the
     * corresponding QueryManipulation.
     *
     * @param obligations which contains all obligations.
     * @return correct obligation.
     */
    public JsonNode getObligation(Iterable<JsonNode> obligations, String queryType) {
        for (JsonNode obligation : obligations) {
            var obligationIsFine = obligationIsFine(obligation, queryType);
            if (obligationIsFine) {
                return obligation;
            } else {
                return NULL_NODE;
            }
        }
        return NULL_NODE;
    }

    /**
     * Checks if an obligation of a {@link io.sapl.api.pdp.Decision} is responsible
     * and can be applied.
     *
     * @param obligations are the obligations of a {@link io.sapl.api.pdp.Decision}
     * @return true if an obligation can be applied.
     */
    public boolean isResponsible(Iterable<JsonNode> obligations, String queryType) {
        for (JsonNode obligation : obligations) {
            var obligationIsFine = obligationIsFine(obligation, queryType);
            if (obligationIsFine) {
                return true;
            }
        }
        return false;
    }

    private boolean obligationIsFine(JsonNode obligation, String queryType) {
        if (!obligation.isNull() && obligation.isObject()) {
            var type = obligation.get(TYPE);
            return (!Objects.isNull(type) && type.isTextual() && queryType.equals(type.asText()));
        }
        return false;
    }
}
